/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author:Ali Erdem Ozcan
 */

package org.objectweb.fractal.cecilia.adl.idl;

import static org.objectweb.fractal.cecilia.adl.idl.util.CUtil.buildInclude;
import static org.objectweb.fractal.cecilia.adl.idl.util.CUtil.buildMethodDefinition;

import java.util.List;
import java.util.Map;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.cecilia.adl.SourceCodeProvider;
import org.objectweb.fractal.cecilia.adl.idl.ast.Field;
import org.objectweb.fractal.cecilia.adl.idl.ast.IDLDefinition;
import org.objectweb.fractal.cecilia.adl.idl.ast.InterfaceDefinition;
import org.objectweb.fractal.cecilia.adl.idl.ast.Method;
import org.objectweb.fractal.task.codegen.util.CodeWriter;
import org.objectweb.fractal.task.core.AbstractTaskFactoryUser;
import org.objectweb.fractal.task.core.Executable;
import org.objectweb.fractal.task.core.TaskException;
import org.objectweb.fractal.task.core.primitive.annotations.ServerInterface;
import org.objectweb.fractal.task.core.primitive.annotations.ServerInterfaces;
import org.objectweb.fractal.task.core.primitive.annotations.TaskParameters;

/**
 * Visitor component for creating virtual function definitions in C from
 * {@link InterfaceDefinition} nodes.
 */
public class CVFTTemplateDefinitionVisitor extends AbstractTaskFactoryUser
    implements
      IDLDefinitionVisitor {

  // ---------------------------------------------------------------------------
  // Implementation of the IDLVisitor interface
  // ---------------------------------------------------------------------------

  /**
   * Visits {@link InterfaceDefiniton} nodes, and creates a task that writes the
   * corresponding C source code.
   */
  public Component visit(final List<Node> path,
      final IDLDefinition idlDefinition, final Map<Object, Object> context)
      throws ADLException, TaskException {
    if (idlDefinition instanceof InterfaceDefinition)
      return taskFactoryItf.newPrimitiveTask(new CVFTTemplateTask(
          (InterfaceDefinition) idlDefinition), idlDefinition);
    else
      return null;
  }

  // ---------------------------------------------------------------------------
  // Task classes
  // ---------------------------------------------------------------------------

  /**
   * Builds virtual function table definitions in C. It extends the Abstract
   * Interface visitor and implements the required visit methods.
   */
  @TaskParameters("interfaceDefinitionNode")
  @ServerInterfaces(@ServerInterface(name = "c-vft-definition-provider", signature = SourceCodeProvider.class, record = "role:cVFTDefinition, id:%", parameters = {"interfaceDefinitionNode"}))
  public static class CVFTTemplateTask extends AbstractInterfaceVisitor
      implements
        Executable,
        SourceCodeProvider {

    protected final InterfaceDefinition itfDefinition;

    // -------------------------------------------------------------------------
    // Task constructor
    // -------------------------------------------------------------------------

    /**
     * @param itfDefinition The interface definition node for which this task
     *            must generate C source code.
     */
    public CVFTTemplateTask(final InterfaceDefinition itfDefinition) {
      this.itfDefinition = itfDefinition;
    }

    // The interface name in C
    protected String itfName;
    // Code writers for different pieces of code,
    protected CodeWriter header, footer, methodDefinition, vft;
    // Final source code that is generated by this task.
    protected String     sourceCode;

    // -------------------------------------------------------------------------
    // Implementation of the Executable interface
    // -------------------------------------------------------------------------

    public void execute() throws Exception {
      final CodeWriter cw = new CodeWriter("C VFT definition builder");
      // Pre-visit operations
      itfName = itfDefinition.getName();
      header = new CodeWriter();
      footer = new CodeWriter();
      methodDefinition = new CodeWriter();
      vft = new CodeWriter();
      // Visit
      visit(itfDefinition);
      // Post-visit operations
      cw.append(header.toString());
      cw.append(methodDefinition.toString());
      cw.append(vft.toString());
      cw.append(footer.toString());
      sourceCode = cw.toString();
    }

    // -------------------------------------------------------------------------
    // Implementation of the SourceCodeProvider interface
    // -------------------------------------------------------------------------

    public String getSourceCode() {
      return sourceCode;
    }

    // --------------------------------------------------------------------------
    // Implementation of the visitor methods
    // --------------------------------------------------------------------------

    public void enterInterface(final InterfaceDefinition itf) throws Exception {
      header.appendln("/* This file is generated automatically */");
      final String cUpItfName = itfName.toUpperCase().replace('.', '_');
      header.append("#ifndef _").append(cUpItfName).append("_VFT_TEMPLATE_H_")
          .endl();
      header.append("#define _").append(cUpItfName).append("_VFT_TEMPLATE_H_")
          .endl();
      header.appendln(buildInclude(itfName));

      footer.appendln("#endif");

      methodDefinition.append("// The following macro can be used to define ")
          .append("prototypes of the methods that a given component must ")
          .append("implements.").endl();
      methodDefinition.append("#define ").append(cUpItfName).append(
          "_METHOD_DEFINITION(comp, itf) \\").endl();

      vft.append(
          "// The following macro can be used to declare and initialize ")
          .append("a VTF instance for a given component interface.").endl();
      vft.append("#define ").append(cUpItfName).append(
          "_VFT_DECLARATION(comp, itf) \\").endl();
      vft.append("  static struct M").append(itfName.replace('.', '_')).append(
          " comp##_##itf##meth = {\\").endl();
    }

    public void leaveInterface(final InterfaceDefinition itf) throws Exception {
      vft.appendln("};");
    }

    public void visitMethod(final Method method) throws Exception {
      // Write VFT initialization for this method
      final String methodName = "comp##_##itf##_##" + method.getName()
          + "##_method";

      vft.append(methodName).append(", \\").endl();

      // Write method definition
      methodDefinition.append(buildMethodDefinition(method, false, methodName))
          .append("; \\").endl();
    }

    public void visitField(final Field field) throws Exception {
      // Nothing to do ...
    }
  }
}
