/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author:Ali Erdem Ozcan
 *
 */

package org.objectweb.fractal.cecilia.composite.c.controllers;

import static org.objectweb.fractal.cecilia.adl.controllers.ControllerInterface.FACTORY;
import static org.objectweb.fractal.cecilia.adl.controllers.ControllerInterface.FACTORY_SIGNATURE;

import java.util.List;
import java.util.Map;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.adl.components.Component;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.adl.interfaces.Interface;
import org.objectweb.fractal.adl.interfaces.InterfaceContainer;
import org.objectweb.fractal.adl.types.TypeInterface;
import org.objectweb.fractal.cecilia.adl.interfaces.InterfaceDecorationUtil;

/**
 * Adds controllers interface for <code>"composite"</code> components.
 */
public class CloneableControllerLoader extends CompositeControllerLoader {

  private static final String CLONEABLE_COMPOSITE = "fractal.lib.cloneableComposite";

  // --------------------------------------------------------------------------
  // Overridden methods
  // --------------------------------------------------------------------------

  @Override
  protected void checkComponent(final ComponentContainer container)
      throws ADLException {
    super.checkComponent(container);

    // checks that each sub component has a factory interface
    for (final Component subComp : container.getComponents()) {
      if (subComp instanceof InterfaceContainer) {
        final TypeInterface subFItf = getInterface(FACTORY,
            ((InterfaceContainer) subComp));
        if (subFItf == null)
          throw new ADLException("A sub component of a Factory composite has "
              + "no \"factory\" interface.", subComp);

        checkServerInterface(subFItf);

        /*
         * set the NO_STATIC_BINDING_DECORATION on the factory-allocator of the
         * sub components.
         */
        final TypeInterface subFAItf = getInterface("factory-allocator",
            ((InterfaceContainer) subComp));
        if (subFAItf != null) {
          InterfaceDecorationUtil.setNoStaticBinding(subFAItf, true);
        }
      } else {
        throw new ADLException("A sub component of a Factory composite is not "
            + "an interface container. Checks the DTD", subComp);
      }
    }
  }

  @Override
  protected List<Interface> getControllerInterfaces(
      final ComponentContainer container, final Map<Object, Object> context)
      throws ADLException {
    final List<Interface> controllerInterfaces = super.getControllerInterfaces(
        container, context);

    final Interface factoryInterface = getFactoryInterface(container, context);
    if (factoryInterface != null) controllerInterfaces.add(factoryInterface);

    final Interface factoryAllocatorInterface = getFactoryAllocatorInterface(
        container, context);
    if (factoryAllocatorInterface != null)
      controllerInterfaces.add(factoryAllocatorInterface);

    return controllerInterfaces;
  }

  @Override
  protected Interface getComponentIdentityInterface(
      final ComponentContainer container, final Map<Object, Object> context)
      throws ADLException {
    final Interface ciItf = super.getComponentIdentityInterface(container,
        context);
    setItfCode(ciItf, CLONEABLE_COMPOSITE, context);
    return ciItf;
  }

  @Override
  protected Interface getBindingControllerInterface(
      final ComponentContainer container, final Map<Object, Object> context)
      throws ADLException {
    final Interface bcItf = super.getBindingControllerInterface(container,
        context);
    setItfCode(bcItf, CLONEABLE_COMPOSITE, context);
    return bcItf;
  }

  @Override
  protected Interface getContentControllerInterface(
      final ComponentContainer container, final Map<Object, Object> context)
      throws ADLException {
    final Interface ccItf = super.getContentControllerInterface(container,
        context);
    setItfCode(ccItf, CLONEABLE_COMPOSITE, context);
    return ccItf;
  }

  @Override
  protected Interface getLifeCycleControllerInterface(
      final ComponentContainer container, final Map<Object, Object> context)
      throws ADLException {
    final Interface lccItf = super.getLifeCycleControllerInterface(container,
        context);
    setItfCode(lccItf, CLONEABLE_COMPOSITE, context);
    return lccItf;
  }

  // --------------------------------------------------------------------------
  // Utility methods
  // --------------------------------------------------------------------------

  protected Interface getFactoryInterface(final ComponentContainer container,
      final Map<Object, Object> context) throws ADLException {
    final InterfaceContainer itfContainer = (InterfaceContainer) container;
    TypeInterface fItf = getInterface(FACTORY, itfContainer);
    if (fItf == null) {
      fItf = newServerInterfaceNode(FACTORY, FACTORY_SIGNATURE);
    } else {
      checkServerInterface(fItf);
      // remove the interface form the interface container, since it is re-added
      itfContainer.removeInterface(fItf);
    }
    setItfCode(fItf, CLONEABLE_COMPOSITE, context);
    return fItf;
  }

  protected Interface getFactoryAllocatorInterface(
      final ComponentContainer container, final Map<Object, Object> context)
      throws ADLException {
    final InterfaceContainer itfContainer = (InterfaceContainer) container;
    // Add the factory-allocator client interface
    TypeInterface allocItf = getInterface("factory-allocator", itfContainer);
    if (allocItf == null) {
      allocItf = newClientInterfaceNode("factory-allocator",
          "memory.api.Allocator");
    } else {
      checkClientInterface(allocItf);
      // remove the interface form the interface container, since it is re-added
      itfContainer.removeInterface(allocItf);
    }
    ((Node) allocItf).astSetDecoration(
        InterfaceDecorationUtil.NO_STATIC_BINDING_DECORATION, true);

    // Adds the CLONEABLE_COMPOSITE code decoration on the "factory-allocator"
    // interface of the composite to ensure that the memory/api/Allocator.idl.h
    // is generated before the compilation of the CLONEABLE_COMPOSITE.
    setItfCode(allocItf, CLONEABLE_COMPOSITE, context);

    return allocItf;
  }

}
