/***
 * Cecilia ADL Compiler
 * Copyright (C) 2005 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.adl.test;

import static org.objectweb.fractal.cecilia.adl.components.ComponentDecorationUtil.getStartOrder;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import junit.framework.TestCase;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.Definition;
import org.objectweb.fractal.adl.JavaCodeLoader;
import org.objectweb.fractal.adl.Loader;
import org.objectweb.fractal.adl.components.Component;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.adl.xml.XMLLoader;
import org.objectweb.fractal.adl.xml.XMLNodeFactoryImpl;
import org.objectweb.fractal.cecilia.adl.bindings.BindingCycleDetectorLoader;
import org.objectweb.fractal.cecilia.adl.bindings.BindingResolverLoader;
import org.objectweb.fractal.cecilia.adl.bindings.StartDependencyCycle;

/**
 * Test the {@link BindingCycleDetectorLoader} component.
 */
public class BindingCycleTest extends TestCase {
  protected static final String ADL_PREFIX = "test.bindings.cycle.";

  protected Loader              l;

  @Override
  protected void setUp() {
    final BindingCycleDetectorLoader cycleLoader = new BindingCycleDetectorLoader();
    final BindingResolverLoader bindLoader = new BindingResolverLoader();
    final XMLLoader xl = new XMLLoader(false);
    xl.nodeFactoryItf = new XMLNodeFactoryImpl();
    bindLoader.interfaceCodeLoaderItf = new JavaCodeLoader();
    bindLoader.clientLoader = xl;
    cycleLoader.clientLoader = bindLoader;
    l = cycleLoader;
  }

  protected Definition load(final String name) throws ADLException {
    return l.load(ADL_PREFIX + name, null);
  }

  protected static void assertCycle(final ADLException e,
      final String[] referenceNames) {
    assertTrue(e instanceof StartDependencyCycle);

    assertNotNull(((StartDependencyCycle) e).getContainer());

    final List<Object> cycle = ((StartDependencyCycle) e).getCycle();
    assertEquals(cycle.size(), referenceNames.length);
    Set<String> nameSet = new HashSet<String>();
    for (final Object o : cycle) {
      assertTrue(o instanceof Component);
      assertTrue(nameSet.add(((Component) o).getName()));
    }
    Set<String> referenceNameSet = new HashSet<String>();
    for (final String n : referenceNames) {
      referenceNameSet.add(n);
    }

    assertEquals(nameSet, referenceNameSet);
  }

  protected static void assertStartOrder(final Definition d,
      final String[][] expectedOrder) {
    List<Component> startOrder = getStartOrder((ComponentContainer) d);
    int expIdx = 0;
    Set<String> expecteds = null;
    for (final Component comp : startOrder) {
      if (expecteds == null || expecteds.isEmpty()) {
        expecteds = new HashSet<String>();
        for (String name : expectedOrder[expIdx])
          expecteds.add(name);
        expIdx++;
      }

      assertTrue(expecteds.remove(comp.getName()));
    }
    assertEquals(expIdx, expectedOrder.length);
  }

  /**
   * Test method.
   * 
   * @throws ADLException
   */
  public void test1() throws ADLException {
    assertStartOrder(load("Test1"), new String[][]{new String[]{"comp4"},
        new String[]{"comp1", "comp2", "comp3"}});

  }

  /**
   * Test method.
   * 
   * @throws ADLException
   */
  public void test2() throws ADLException {
    assertStartOrder(load("Test2"), new String[][]{
        new String[]{"comp3", "comp1"}, new String[]{"comp2"}});
  }

  /**
   * Test method.
   */
  public void test3() {
    try {
      load("Test3");
      fail("An exception was expected");
    } catch (final ADLException e) {
      assertCycle(e, new String[]{"comp1", "comp2"});
    }
  }

  /**
   * Test method.
   */
  public void test4() {
    try {
      load("Test4");
      fail("An exception was expected");
    } catch (final ADLException e) {
      assertCycle(e, new String[]{"comp1", "comp2"});
    }
  }

  /**
   * Test method.
   */
  public void test5() {
    try {
      load("Test5");
      fail("An exception was expected");
    } catch (final ADLException e) {
      assertCycle(e, new String[]{"comp1", "comp2"});
    }
  }

  /**
   * Test method.
   */
  public void test6() {
    try {
      load("Test6");
      fail("An exception was expected");
    } catch (final ADLException e) {
      assertCycle(e, new String[]{"comp1", "comp3"});
    }
  }

  /**
   * Test method.
   * 
   * @throws ADLException
   */
  public void test7() throws ADLException {
    assertStartOrder(load("Test7"),
        new String[][]{new String[]{"comp2", "comp3", "comp4", "comp5"},
            new String[]{"comp1"}});
  }

  /**
   * Test method.
   * 
   * @throws ADLException
   */
  public void test8() throws ADLException {
    assertStartOrder(load("Test8"), new String[][]{
        new String[]{"comp4", "comp5"}, new String[]{"comp3"},
        new String[]{"comp2"}, new String[]{"comp1"}});
  }
}
