/**
 Copyright 2007 INRIA
 
 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 USA
 
 Contact: fractal@objectweb.org
 Author:  Alessio Pace
 */

package org.objectweb.fractal.cecilia.adl.maven.plugin;

import java.io.File;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.project.MavenProject;

/**
 * A simple mojo which allows invoking the <strong>Cecilia ADL Toolchain</strong>
 * passing some configuration.
 * 
 * The invocation will happen in the same VM which runs the MOJO.
 * 
 * @goal compile
 * 
 * @author Alessio Pace
 */
public class CeciliaAdlMojo extends AbstractMojo {

  /**
   * The Maven project reference.
   * 
   * @parameter expression="${project}"
   * @required
   */
  private MavenProject project;

  /**
   * The fully qualified name of the component ADL definition to compile.
   * 
   * @parameter
   */
  private String adl;

  /**
   * The list of fully qualified name of the component ADL definitions to
   * compile.
   * 
   * @parameter
   */
  private List adls;

  /**
   * The output directory for the files produced by Cecilia ADL.
   * 
   * @parameter expression="${project.build.directory}/build"
   */
  private String outDir;

  /**
   * @parameter
   */
  private List arguments;

  protected ClassLoader getClassLoader(ClassLoader parent) {

    String outputDirectory = project.getBuild().getOutputDirectory();
    try {
      URL targetClassesUrl = new File(outputDirectory).toURI().toURL();

      URLClassLoader cl = new URLClassLoader(new URL[] { targetClassesUrl },
          parent);
      return cl;
    } catch (Exception e) {
      e.printStackTrace();
      throw new RuntimeException("Unexpected exception");
    }

  }

  public void execute() throws MojoExecutionException, MojoFailureException {
    List allOrderedArguments = new ArrayList();

    if (adl != null) {
      if (adls != null)
        throw new RuntimeException("<adl> and <adls> tags cannot be used in the same time.");
      getLog().debug("Compiling architecture: " + adl);
      allOrderedArguments.add(adl);
    } else {
      if (adls == null)
        throw new RuntimeException("At least one <adl> or <adls> tag must be specified.");
      getLog().debug("Compiling architecture: " + adls);
      /* the first argument is the ADL to be compiled */
      Iterator adlIterator = adls.iterator();
      while (adlIterator.hasNext()) {
        CeciliaAdlLauncherArguments adlName = (CeciliaAdlLauncherArguments) adlIterator
            .next();
        allOrderedArguments.addAll(adlName.getArguments(project));
      }
    }

    getLog().debug("Properties " + arguments);

    /*
     * add by default the project build outputdirectory in the toolchain
     * src-path (in this way the project .c and .idl will be found.
     */
    allOrderedArguments.add("-src-path="
        + this.project.getBuild().getOutputDirectory());

    /* add the "-out-dir" parameter also */
    allOrderedArguments.add("-out-dir=" + this.outDir);

    /* creates the output dir if necessary */
    OutDirUtil.createOutDirIfNecessary(this.outDir);

    /* append the other parameters */
    Iterator iterator = arguments.iterator();
    while (iterator.hasNext()) {
      CeciliaAdlLauncherArguments arguments = (CeciliaAdlLauncherArguments) iterator
          .next();
      allOrderedArguments.addAll(arguments.getArguments(project));
    }

    final String[] args = (String[]) allOrderedArguments.toArray(new String[0]);
    getLog().debug(
        "Invoking Cecilia ADL Launcher with the following command line parameters: "
            + Arrays.deepToString(args));

    ClassLoader initialClassLoader = Thread.currentThread()
        .getContextClassLoader();

    ClassLoader newClassLoader = getClassLoader(initialClassLoader);

    Thread.currentThread().setContextClassLoader(newClassLoader);

    try {
      Class launcher = newClassLoader
          .loadClass("org.objectweb.fractal.cecilia.adl.Launcher");
      Method main = launcher.getMethod("main", new Class[] { String[].class });
      main.invoke(launcher, new Object[] { args });
    } catch (Exception e) {
      e.printStackTrace();
      throw new RuntimeException(
          "Unable to obtain or call Cecilia ADL Launcher", e);
    }

    Thread.currentThread().setContextClassLoader(initialClassLoader);

  }

}
